import os
import asyncio
import time
from typing import Optional
from abc import ABC, abstractmethod

from patent_chart import settings
from patent_chart.redis_utils import r, AsyncRedisLock, async_redis

TTL_KEY_DOESNT_EXIST = -2

Lock = asyncio.Lock | AsyncRedisLock

class FixedWindowRateLimiterBase(ABC):
    def __init__(self, capacity: int, fill_interval: float | int, lock: Lock):
        self.capacity = capacity
        self.fill_interval = fill_interval
        self.lock = lock

    @property
    @abstractmethod
    def tokens(self):
        pass
    
    @tokens.setter
    @abstractmethod
    def tokens(self, value):
        pass

    @property
    @abstractmethod
    def last_fill_time(self):
        pass

    @last_fill_time.setter
    @abstractmethod
    def last_fill_time(self, value):
        pass

    def close(self):
        self.refill_task.cancel()

    async def remove_tokens(self, n: int):
        async with self.lock:
            self.tokens -= n
    
    async def request_and_remove_tokens(self, n: int):
        async with self.lock:
            if self.tokens >= n:
                self.tokens -= n
                return True
            else:
                return False
            
    async def wait_for_tokens(self, n: int, timeout: int = 3600):
        if n > self.capacity:
            raise ValueError(f"n tokens {n} must be less than or equal to capacity {self.capacity}")
        start_time = time.monotonic()
        while True:
            if await self.request_and_remove_tokens(n):
                return True
            if time.monotonic() - start_time > timeout:
                return False
            await asyncio.sleep(self.fill_interval - (time.monotonic() - self.last_fill_time))

    async def fill_tokens(self):
        async with self.lock:
            self.tokens = self.capacity
            self.last_fill_time = time.monotonic()

    async def refill(self):
        while True:
            await self.fill_tokens()
            await asyncio.sleep(self.fill_interval)

    async def __aenter__(self):
        self.refill_task = asyncio.create_task(self.refill())
        return self
    
    async def __aexit__(self, exc_type, exc_value, traceback):
        self.close()
        return False

class FixedWindowRateLimiter(FixedWindowRateLimiterBase):
    def __init__(self, capacity: int, fill_interval: float):
        super().__init__(capacity, fill_interval, asyncio.Lock())

    @property
    def last_fill_time(self):
        return self._last_fill_time
    
    @last_fill_time.setter
    def last_fill_time(self, value):
        self._last_fill_time = value

    @property
    def tokens(self):
        return self._tokens
    
    @tokens.setter
    def tokens(self, value):
        self._tokens = value
    

class DistributedFixedWindowRateLimiter(FixedWindowRateLimiterBase):
    def __init__(self, capacity: int, fill_interval: int):
        assert fill_interval > 0, "fill_interval must be greater than 0"
        self.async_r = async_redis.Redis(host=settings.REDIS_HOST, port=6379)
        openai_rate_limit_lock = async_redis.lock.Lock(self.async_r, settings.REDIS_OPENAI_RATE_LIMIT_LOCK_KEY, timeout=3600)
        super().__init__(capacity, fill_interval, AsyncRedisLock(openai_rate_limit_lock))
        
    @property
    def last_fill_time(self):
        last_fill_time = r.get(settings.REDIS_OPENAI_RATE_LIMIT_LAST_FILL_TIME_KEY)
        if last_fill_time is None:
            return -1
        return float(last_fill_time)
    
    @last_fill_time.setter
    def last_fill_time(self, value):
        exp = int(self.fill_interval * 2)
        r.setex(settings.REDIS_OPENAI_RATE_LIMIT_LAST_FILL_TIME_KEY, exp, value)

    @property
    def tokens(self):
        tokens = r.get(settings.REDIS_OPENAI_RATE_LIMIT_TOKENS_KEY)
        if tokens is None:
            return 0
        return int(tokens)
    
    @tokens.setter
    def tokens(self, value):
        ttl = r.ttl(settings.REDIS_OPENAI_RATE_LIMIT_TOKENS_KEY)
        if ttl <= 0:
            ttl = self.fill_interval
        r.setex(settings.REDIS_OPENAI_RATE_LIMIT_TOKENS_KEY, ttl, value)

    async def fill_tokens(self):
        async with self.lock:
            tokens_ttl = r.ttl(settings.REDIS_OPENAI_RATE_LIMIT_TOKENS_KEY)
            if tokens_ttl == TTL_KEY_DOESNT_EXIST:
                try:
                    result = await self.async_r.setex(settings.REDIS_OPENAI_RATE_LIMIT_TOKENS_KEY, self.fill_interval, self.capacity)
                except Exception as e:
                    print(e)
                self.last_fill_time = time.monotonic()

    async def refill(self):
        while True:
            await self.fill_tokens()
            # sleep for the remaining time in the fill interval + 1 second, then whichever process wakes up first will refill the tokens
            sleep_for = self.fill_interval - (time.monotonic() - self.last_fill_time) + 1
            await asyncio.sleep(sleep_for)

    async def __aexit__(self, exc_type, exc_value, traceback):
        super().close()
        await self.async_r.aclose()
        return False


import xml.etree.ElementTree as ET
import xml.dom.minidom as minidom

from openai import OpenAI
from langchain.text_splitter import RecursiveCharacterTextSplitter

from .parser import PatentUniqueID
from . import search_index

MAX_SEQ_LENGTH = 8191
MAX_BATCH_SIZE=2048

oai = OpenAI()


def prettify(elem):
    """Return a pretty-printed XML string for the Element."""
    rough_string = ET.tostring(elem, 'utf-8')
    reparsed = minidom.parseString(rough_string)
    return reparsed.toprettyxml(indent="  ")


def extract_patent_unique_id(xml: ET.Element) -> PatentUniqueID | None:
    """
    Extract unique id given patent xml root
    """
    doc_id = xml.find('.//us-bibliographic-data-grant/publication-reference/document-id')
    if doc_id is None:
        return None
    try:
        country_code = doc_id.find('.//country').text
        patent_number = doc_id.find('.//doc-number').text
    except AttributeError:
        return None
    # For some reason some patent numbers have leading zeros
    patent_number = patent_number.lstrip('0')
    kind_code = doc_id.find('.//kind')
    if kind_code is not None:
        kind_code = kind_code.text
    return PatentUniqueID(patent_number=patent_number, country_code=country_code, kind_code=kind_code)


def extract_application_filing_date(xml: ET.Element) -> str:
    """
    Extract application filing date given patent xml root
    """
    app_filing_date = xml.find('.//us-bibliographic-data-grant/application-reference/document-id/date')
    return f'{app_filing_date.text[:4]}-{app_filing_date.text[4:6]}-{app_filing_date.text[6:]}'


def extract_priority_date(xml: ET.Element) -> str | None:
    """
    Extract priority date given patent xml root returns in YYYY-MM-DD
    """
    priority_claim = xml.find(".//priority-claim[@sequence='01']")
    
    if priority_claim is not None:
        date_element = priority_claim.find("date")
        if date_element is not None:
            return f'{date_element.text[:4]}-{date_element.text[4:6]}-{date_element.text[6:]}'
    
    provisional_application = xml.find(".//us-provisional-application")
    if provisional_application is not None:
        date_element = provisional_application.find(".//date")
        if date_element is not None:
            return f'{date_element.text[:4]}-{date_element.text[4:6]}-{date_element.text[6:]}'
    
    us_related_docs = xml.find(".//us-related-documents")
    if us_related_docs is not None:
        parent_date = us_related_docs.find(".//continuation/relation/parent-doc/document-id/date")
        if parent_date is not None:
            return f'{parent_date.text[:4]}-{parent_date.text[4:6]}-{parent_date.text[6:]}'
        
    return None


def extract_title(xml):
    title = xml.find('.//invention-title')
    if title is None:
        return ''
    return title.text


def extract_claims(xml: ET.Element):
    claims = xml.find('.//claims')
    return claims


def extract_written_description(xml: ET.Element):
    description = xml.find('.//description')
    return description


def extract_written_description_text(description: ET.Element) -> str:
    text = ''
    for elem in description:
        if elem.tag not in ['p', 'heading']:
            elem_text = '\n'.join([extract_text_recursive(child) for child in elem])
        else:
            elem_text = extract_text_recursive(elem)
        text += elem_text + '\n'
        if elem.tag == 'p':
            text += '\n'
    return text


def extract_text_recursive(element: ET.Element):
    text = ''
    if element.text:
        text += element.text
    for child in element:
        text += extract_text_recursive(child)
    if element.tail:
        text += element.tail
    return text


def chunk_claims(claims: ET.Element) -> list[str]:
    claims_texts = []

    for claim in claims:
        text = extract_text_recursive(claim).strip()
        claims_texts.append(text[:MAX_SEQ_LENGTH])  # HACK: for now, just truncating

    return claims_texts


def chunk_written_description(description: ET.Element, chunk_size_words=512, chunk_overlap_words=128) -> list[str]:
    text = ''
    for elem in description:
        if elem.tag not in ['p', 'heading']:
            elem_text = ' '.join([extract_text_recursive(child) for child in elem])
        else:
            elem_text = extract_text_recursive(elem)
        text += elem_text
        if elem.tag == 'p':
            text += '\n'

    text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(
        model_name='gpt-4',
        chunk_size=chunk_size_words,
        chunk_overlap=chunk_overlap_words,
    )

    chunks = text_splitter.split_text(text)
    return chunks


def chunk_written_description_with_pos(description: ET.Element, chunk_size_words=8191, chunk_overlap_words=1024) -> tuple[list[str], list[tuple[int]]]:
    text = ''
    for elem in description:
        if elem.tag not in ['p', 'heading']:
            elem_text = ' '.join([extract_text_recursive(child) for child in elem])
        else:
            elem_text = extract_text_recursive(elem)
        text += elem_text
        if elem.tag == 'p':
            text += '\n'

    return chunk_written_description_from_text_with_pos(text, chunk_size_words, chunk_overlap_words)


def chunk_written_description_from_text_with_pos(text: str, chunk_size_words=8191, chunk_overlap_words=1024) -> tuple[list[str], list[tuple[int]]]:
    text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(
        model_name='gpt-4',
        chunk_size=chunk_size_words,
        chunk_overlap=chunk_overlap_words,
    )

    chunks = text_splitter.split_text(text)

    # Find start and end char pos for each chunk
    pos = 0
    starts = []
    ends = []
    for chunk in chunks:
        starts.append(pos)
        pos += len(chunk)
        ends.append(pos)

    return chunks, list(zip(starts, ends))


def should_drop(patent_unique_id: PatentUniqueID) -> bool:
    """
    Determine if the xml should be dropped
    """
    if patent_unique_id and patent_unique_id.kind_code not in ['B1', 'B2', 'A1', 'A2', 'A9']:
        return True
    return False


def inclusion(data: str, super_string: str) -> bool:
    return data in super_string


def min_length(data: str, min_length: int) -> bool:
    return len(data) >= min_length


def get_filter_mask(data: list[str], *fns):
    mask = [True] * len(data)
    for fn in fns:
        mask = [x and fn(data[i]) for i, x in enumerate(mask)]
      
    return mask

def apply_filter_pipeline(data: list[str], *fns):
    mask = get_filter_mask(data, *fns)
    return [data[i] for i, x in enumerate(mask) if x]


from dataclasses import dataclass
from typing import Optional

import pydantic

@dataclass
class PatentTextLine:
    page_num: int  # 0-based index returned from PDFPageDetailedAggregator. TODO: rename to index
    text: str
    line_index: int  # line index counting TextLine instances from top to bottom of page
    column_index: Optional[int]  # 0 or 1 or None if noncolumnar
    line_num: Optional[int] = None  # human labeled line number, multiple of 5
    column_num: Optional[int] = None  # human labeled column number
    paragraph_num: Optional[int] = None  # alternative human labeling method often in form [XXXX] at beginning of paragraph

@dataclass
class GeneratedPassage:
    prior_art_source_id: int
    text: str
    claim_element_id: int
    model_id: str
    start_line: Optional[PatentTextLine] = None
    end_line: Optional[PatentTextLine] = None
    ranking: int = 0


@dataclass
class PatentUniqueID:
    patent_number: str
    country_code: str
    kind_code: Optional[str] = None

    def __repr__(self):
        return f'{self.country_code}{self.patent_number}'

    def __str__(self):
        return self.__repr__()

    @classmethod
    def from_patent_id(cls, patent_id: str):
        country_code = patent_id[:2]
        patent_number = patent_id[2:]
        return cls(patent_number=patent_number, country_code='US')

@dataclass
class GoogleClaim:
    claim_number: int
    claim_elements: list[str]


class GoogleParsedPatent(pydantic.BaseModel):
    unique_id: PatentUniqueID
    specification: Optional[str] = None
    claims: Optional[str] = None
    title: Optional[str] = None
    text: Optional[str] = None
    text_format: Optional[str] = None
    filing_date: Optional[str] = None
    priority_date: Optional[str] = None


import re
import os
import time
import asyncio
import uuid
import random
from logging import getLogger
from dataclasses import dataclass
from typing import Callable, Optional, AsyncIterator, Coroutine, List
from functools import partial, lru_cache

from openai import OpenAI, AsyncOpenAI, OpenAIError
import tiktoken
from anthropic import AsyncAnthropic
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
)

from . import parser
from . import ranking
from . import search_index
from .utils import num_tokens_from_messages
from .data_structures import GeneratedPassage
from .requests_utils import DistributedFixedWindowRateLimiter

logger = getLogger(__name__)

oai = OpenAI()
async_oai = AsyncOpenAI()

ac = AsyncAnthropic()

class GenerationError(Exception):
    pass

RATE_LIMIT_INTERVAL = 60  # seconds

# OpenAI text-embedding-3-large
MAX_SEQ_LENGTH = 8191

# rate limit
token_rate_limit_by_model_name = {
    'gpt-4': 300_000,
    'gpt-3.5-turbo': 2_000_000,
    'gpt-3.5-turbo-16k': 2_000_000,
    'gpt-3.5-turbo-instruct': 250_000,
    'gpt-3.5-turbo-0125': 2_000_000,
    'gpt-4-turbo-preview': 600_000,
    'claude-3-opus-20240229': 400_000,
}

# context window
model_name_to_token_limit = {
    'gpt-4': 8192,
    'gpt-3.5-turbo-16k': 16384,
    'gpt-3.5-turbo': 4096,
    'gpt-3.5-turbo-16k-0613': 16384,
    'gpt-3.5-turbo-0125': 16385,
    'gpt-4-turbo-preview': 128_000,
    'claude-3-opus-20240229': 200_000,
}


def prompt_claim_element_only(claim_element, prior_art_passage, n_passages_per_split=3):
    messages = [
        {"role": "system", "content": f"You work for a law firm. Your task is to interpret a single claim element that I select from the claims of a patent. Then I'm going to give you a passage from the prior art and you need to pick {n_passages_per_split} distinct sub-passages from the passage that you believe are the most similar to the claim element. Please output only those {n_passages_per_split} passages, copied verbatim from the prior art. Do not add any other text besides the verbatim passages."},
        {"role": "user", "content": f"Selected claim element: {claim_element}"},
        {"role": "user", "content": f"Prior art passage: {prior_art_passage}"},
    ]
    return messages

def prompt_summary_full_set_of_claims(summary, claims, claim_element, prior_art_passage, n_passages_per_split=3):
    messages = [
        {"role": "system", "content": f"You work for a law firm. Your task is to read the summary of the written description of a patent and the full set of claims in the patent and interpret a single claim element that I select from the claims of the patent in light of the summary and the full set of claims. Then I'm going to give you a passage from the prior art and you need to pick{n_passages_per_split} distinct sub-passages from the passage that you believe are the most similar to the claim element. Please output only those {n_passages_per_split} passages, copied verbatim from the prior art, each separated by a single newline. Do not add any other text besides the verbatim passages. If you don't believe there are any sub-passages that are similar to the claim element, please output 'None'."},
        {"role": "user", "content": f"Summary of the written description of the invention: {summary}\n\nFull set of claims: {claims}\n\nSelected claim element: {claim_element}Prior art passage: {prior_art_passage}"},
    ]
    return messages


def prompt_full_spec_full_set_of_claims(full_spec, claims, claim_element, prior_art_passage, n_passages_per_split=3):
    messages = [
        {"role": "system", "content": f"You work for a law firm. Your task is to read the the written description of a patent and the full set of claims in the patent and interpret a single claim element selected from the claims of the patent in the context of the written description and the other claim elements. Then I'm going to give you a prior art document and you need to pick up to {n_passages_per_split} distinct passages from the prior art document that you believe are the most similar to the claim element. Please output only those {n_passages_per_split} passages, copied verbatim from the prior art document, each separated by a single '\n'. Do not add any other text besides the verbatim passages. If you don't believe there are any passages that are similar to the claim element, please output 'None'."},
        {"role": "user", "content": f"Written description from the patent: {full_spec} \n\nFull set of claims from the patent: {claims} \n\nSelected claim element from the claims of the patent: {claim_element} \n\nPrior art document: {prior_art_passage}"},
    ]
    return messages


def post_process_selected_passage_gpt4(passage: str) -> list[str]:
    split_strategies = [
        lambda passage: re.split(r'\d+\.', passage),
        lambda passage: passage.split('\n\n'),
        lambda passage: passage.split('\"\n'),
    ]
    for split_strategy in split_strategies:
        split_passages = split_strategy(passage)
        if len(split_passages) > 1:
            break
    # strip newlines, quotes, and whitespace from beginning and end
    split_passages = [p.strip().strip('"') for p in split_passages]
    # filter whitespace only and empty string entries
    split_passages = [p for p in split_passages if p != '']
    # filter out passages that are just a newline
    split_passages = [p for p in split_passages if p != '\n']
        
    return split_passages


def post_process_selected_passage_claude_opus(passage: str) -> list[str]:
    passage = passage.split(':')[-1]
    
    split_strategies = [
        lambda passage: passage.split('\n'),
    ]
    for split_strategy in split_strategies:
        split_passages = split_strategy(passage)
        if len(split_passages) > 1:
            break
    # strip newlines, quotes, and whitespace from beginning and end
    split_passages = [p.strip().strip('"') for p in split_passages]
    # filter whitespace only and empty string entries
    split_passages = [p for p in split_passages if p != '']
    # filter out passages that are just a newline
    split_passages = [p for p in split_passages if p != '\n']
    
    return split_passages
  

def post_process_summary_claude_opus(summary: str) -> str:
    summary = summary.split(':')[-1]
    return summary.strip().strip('"')


@dataclass
class GeneratorConfig:
    model_name: str = 'gpt-4'
    target_n_passages: int = 10
    min_n_splits: int = 1

def load_generator_config_from_env():
    model_name = os.environ.get('LLM_MODEL_NAME', 'gpt-4')
    if model_name not in token_rate_limit_by_model_name:
        raise ValueError(f"Model name {model_name} not recognized.")
    target_n_passages = int(os.environ.get('LLM_TARGET_N_PASSAGES', 10))
    min_n_splits = int(os.environ.get('LLM_MIN_N_SPLITS', 1))
    
    logger.info('Loaded generator config: \n %s \n', )

    config = GeneratorConfig(
        model_name=model_name,
        target_n_passages=target_n_passages,
        min_n_splits=min_n_splits,
    )

    logger.info('Loaded generator config: \n %s \n', config)

    return config

def num_tokens_in_text(text, model="gpt-3.5-turbo-0613"):
  """Returns the number of tokens in a string."""
  try:
      encoding = tiktoken.encoding_for_model(model)
  except KeyError:
      encoding = tiktoken.get_encoding("cl100k_base")
  return len(encoding.encode(text))


def concatenate_completions_for_reranking(completions):
    # concatenate completions and renumber
    # TODO: make 3 a parameter
    concatenated_completions = ''
    for i, completion in enumerate(completions):
        completion_content = completion['choices'][0]['message']['content']
        completion_content1, completion_content = completion_content.split('\n2.')
        completion_content1 = completion_content1.lstrip('1. ')
        completion_content2, completion_content = completion_content.split('\n3.')
        completion_content3 = completion_content
        
        concatenated_completions += f'{int_to_roman(i*3 + 1)}. {completion_content1}\n'
        concatenated_completions += f'{int_to_roman(i*3 + 2)}. {completion_content2}\n'
        concatenated_completions += f'{int_to_roman(i*3 + 3)}. {completion_content3}\n'
    return concatenated_completions, 3 * len(completions)

# Reranking utility functions
def int_to_roman(n: int) -> str:
    if not (0 < n < 4000):
        raise ValueError("Input integer must be between 1 and 3999")
    
    ints = (1000, 900,  500, 400, 100,  90, 50,  40, 10,  9,   5,  4,   1)
    nums = ('M',  'CM', 'D', 'CD', 'C', 'XC','L','XL','X','IX','V','IV','I')
    result = []
    
    for i in range(len(ints)):
        count = int(n / ints[i])
        result.append(nums[i] * count)
        n -= ints[i] * count
    
    return ''.join(result)

# parse reranked completions
def parse_reranked_completions(reranked_completion, n_completions):
    # First lstrip prelude before 1.
    # Find index of '1. ' and slice
    reranked_completion = reranked_completion[reranked_completion.index('1. '):]
    
    # Split on i. for i in range(1, n_completions + 1)
    # Then lstrip roman numeral and period
    roman_numeral_regex_pattern = r''
    for i in range(1, n_completions + 1):
        roman_numeral_regex_pattern += f'{int_to_roman(i)}\. '
        if i < n_completions:
            roman_numeral_regex_pattern += '|'
    roman_numeral_regex = re.compile(roman_numeral_regex_pattern)
    
    parsed_completions = []
    for i in range(1, n_completions + 1):
        if i == n_completions:
            parsed_completion = reranked_completion
        else:
            parsed_completion, reranked_completion = reranked_completion.split(f'{i+1}. ', maxsplit=1)
        if i == 1:
            parsed_completion = parsed_completion.lstrip('1. ')
        parsed_completion = re.sub(roman_numeral_regex, '', parsed_completion)
        parsed_completion = parsed_completion.lstrip(' "')
        parsed_completion = parsed_completion.rstrip(' \n"')
        parsed_completions.append(parsed_completion)
        
    return parsed_completions


def openai_prompt_select_passages_from_prior_art_portion(all_but_claims, claims, claim_element, prior_art_passage):
    # TODO: make 3 a parameter
    messages = [
        {"role": "system", "content": "You are an associate at a big law firm. Your task is to read the written description and of an invention disclosed in a patent and the full set of claims in the patent and interpret a single claim element that I select from the claims of the patent in light of the written description and the full set of claims. Then I'm going to give you a passage from of prior art and you need to pick three distinct sub-passages from the passage that you believe are the most semantically similar to the claim element interpreted in light of the written description and the full set of claims. Please output only those three passages, each preceded only by the numberings 1., 2., and 3. and a space."},
        {"role": "user", "content": f"Written description of invention: {all_but_claims}"},
        {"role": "user", "content": f"Full set of claims: {claims}"},
        {"role": "user", "content": f"Selected claim element: {claim_element}"},
        {"role": "user", "content": f"Prior art passage: {prior_art_passage}"},
    ]
    return messages

def openai_prompt_rank_selected_passages(all_but_claims, claims, claim_element, chosen_prior_art_passage, n_completions):
    messages = [
        {"role": "system", "content": f"Your task is to read the written description and of an invention disclosed in a patent and the full set of claims in the patent and interpret a single claim element that I select from the claims of the patent in light of the written description and the full set of claims. Then I'm going to give you {n_completions} sub-passages from the prior art passage i presented to you that you chose because they were the most semantically similar to the selected claim element interpreted in light of the written description and the full set of claims. Please rank these {n_completions} passages in descending order of similarity to the selected claim element interpreted in light of the written description and the full set of claims. I will give you passages numbered by roman numerals, for example: 'I. Passage One... II. Passage two... III. Passage three...'. Please output only the integer ranking following by the roman numeral and passage, for example: '1. II. Passage two ... 2. I. Passage one... 3. III Passage three...' if you believe the ranking should be passage two > passage one > passage three."},
        {"role": "user", "content": f"Written description of invention: {all_but_claims}"},
        {"role": "user", "content": f"Claim element: {claim_element}"},
        {"role": "user", "content": f"Prior art passages to rank: {chosen_prior_art_passage}"},
    ]
    return messages

def openai_chat_completion_request_with_retry(model_name, messages, backoff_factor=2, backoff_override=None) -> str:
    for i in range(1, 3):
        try:
            completion = oai.chat.completions.create(
                model=model_name,
                messages=messages
            )
            return completion.choices[0].message.content
        except OpenAIError as e:
            if backoff_override is not None:
                time.sleep(backoff_override)
            else:
                time.sleep(backoff_factor**i)
            continue
        except Exception as e:
            raise e
    raise Exception("Failed to send chat completion request to OpenAI after 4 retries.")


async def aopenai_chat_completion_request_with_retry(model_name, messages, backoff_factor=2, backoff_override=None, n_retries=3, rate_limiter: Optional[DistributedFixedWindowRateLimiter] = None, **kwargs) -> str:
    """Send a chat completion request to OpenAI with retries."""
    for i in range(1, 1 + n_retries):
        try:
            if rate_limiter is not None:
                num_tokens_in_messages = num_tokens_from_messages(messages)
                # NOTE: separating wait for tokens, the request, and removing tokens is prone to race conditions. We can tolerate occasionally making requests that receive rate limit errors so long as we retry them with backoff.
                if await rate_limiter.wait_for_tokens(num_tokens_in_messages):
                    completion = await async_oai.chat.completions.create(
                        model=model_name,
                        messages=messages,
                        **kwargs
                    )
                    n_tokens_in_completion = completion.usage.completion_tokens
                    await rate_limiter.remove_tokens(n_tokens_in_completion)
            else:
                completion = await async_oai.chat.completions.create(
                    model=model_name,
                    messages=messages,
                    **kwargs
                )
            return completion.choices[0].message.content
        except OpenAIError as e:
            logger.error(f"OpenAIError: {e}")
            if backoff_override is not None:
                jitter = random.uniform(-0.10, 0.10)
                await asyncio.sleep(backoff_override * (1 + jitter))
            else:
                jitter = random.uniform(-0.10, 0.10) 
                await asyncio.sleep(backoff_factor**i * (1 + jitter))
            continue
        except Exception as e:
            raise e
    raise Exception(f"Failed to send chat completion request to OpenAI after {n_retries} retries.")


def split_passage_to_meet_model_token_limit(passage: str, prompt_func: Callable[[str], list[dict]], model_name: str, min_n_splits: int = 1) -> list[str]:
    """ 
    Given a passage string, a prompt func that takes only the passage string, and a model, find the smallest number of splits necessary to get messages under token limit and return splits of the passage.
    """
    token_limit = model_name_to_token_limit[model_name]
    min_split_size = 256
    # We use cl100k_base as the tokenizer for estimating even for Claude which has a closed-source tokenizer.
    base_num_tokens = num_tokens_from_messages(
        prompt_func(' ' * min_split_size)
    )
    if base_num_tokens > token_limit:
        raise GenerationError("Prompt exceeds token limit by itself.")
    
    passage_splits = []
    passage_split_count = min_n_splits
    while True:
        # Check all splits
        candidate_splits = []
        for i in range(passage_split_count):
            if i == 0:
                candidate_split = passage[:len(passage) // passage_split_count]
            else:
                candidate_split = passage[len(passage) // passage_split_count * i:len(passage) // passage_split_count * (i+1)]
            candidate_splits.append(candidate_split)
        num_tokens = [num_tokens_from_messages(prompt_func(split)) for split in candidate_splits]
        if all([n <= token_limit - 0.05 * token_limit for n in num_tokens]):
            passage_splits = candidate_splits
            break
        passage_split_count += 1

    return passage_splits


@retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6), reraise=True)
async def aclaude_chat_completion_request(model_name, messages) -> str:
    system_message = messages.pop(0)
    completion = await ac.messages.create(
        max_tokens=4096,
        system=system_message['content'],
        messages=messages,
        model=model_name,
    )
    print(completion)
    return completion.content[0].text


def make_get_completion_coro(model_name, messages, rate_limiter=None):
    if model_name == 'gpt-4':
        return aopenai_chat_completion_request_with_retry(model_name, messages, rate_limiter=rate_limiter)
    elif model_name == 'claude-3-opus-20240229':
        return aclaude_chat_completion_request(model_name, messages)


async def agenerate_passages_for_prior_art_splits(prior_art_splits: list[str], prompt_func: Callable[[str], list[dict]], model_name: str, claim_element_id: int, prior_art_id: int, rate_limiter: Optional[DistributedFixedWindowRateLimiter] = None) -> list[GeneratedPassage | Exception]:
    # make parallel requests to openai api for prior art splits
    passages: list[GeneratedPassage | Exception] = []
    coros = []
    for prior_art_split in prior_art_splits:
        messages = prompt_func(prior_art_split)
        coros.append(make_get_completion_coro(model_name, messages, rate_limiter=rate_limiter))
        
    completions = await asyncio.gather(*coros, return_exceptions=True)
    for i, completion in enumerate(completions):
        if isinstance(completion, Exception):
            print(f'Failed to get completion for prior art split {i} of prior art {prior_art_id} for claim element {claim_element_id}')
            print(completion)
            passages.append(completion)
        else:
            selected_passages = completion
            if model_name == 'gpt-4':
                post_process_generated_passages_func = post_process_selected_passage_gpt4
            elif model_name == 'claude-3-opus-20240229':
                post_process_generated_passages_func = post_process_selected_passage_claude_opus
            else:
                raise NotImplementedError(f"Post processing function for model {model_name} not implemented.")
            
            post_processed_passages = post_process_generated_passages_func(selected_passages)

            for processed_passage in post_processed_passages:
                # We'll add the start line and end line later
                passages.append(
                    GeneratedPassage(
                        prior_art_source_id=prior_art_id,
                        text=processed_passage,
                        claim_element_id=claim_element_id,
                        model_id=model_name
                    )
                )

    passages = [p for p in passages if not isinstance(p, Exception)]

    return passages


async def agenerate_passages(claim_elements: list[tuple[int, str]], prior_art: tuple[int, str], prompt_func: Callable[[str], list[dict]], model_name: str, target_n_passages: int = 10, min_n_splits:int = 1, rate_limiter: Optional[DistributedFixedWindowRateLimiter] = None) -> AsyncIterator[tuple[int, tuple[int, list[str]]]]:
    tasks = []
    for i, claim_element in claim_elements:
        prior_art_splits = split_passage_to_meet_model_token_limit(prior_art[1], partial(prompt_func, claim_element), model_name, min_n_splits=min_n_splits)
        n_passages_per_split = max(target_n_passages // len(prior_art_splits), 1)
        tasks.append(asyncio.create_task(
            agenerate_passages_for_prior_art_splits(prior_art_splits, partial(prompt_func, claim_element, n_passages_per_split=n_passages_per_split), model_name, claim_element_id=i, prior_art_id=prior_art[0], rate_limiter=rate_limiter)
        ))
    
    for coro in asyncio.as_completed(tasks):
        yield await coro


def create_generate_passages_tasks(claim_elements: list[tuple[int, str]], prior_art: tuple[int, str], prompt_func: Callable[[str], list[dict]], model_name: str, target_n_passages: int = 10, min_n_splits:int = 1, rate_limiter: Optional[DistributedFixedWindowRateLimiter] = None) -> list[asyncio.Task]:
    tasks = []
    for i, claim_element in claim_elements:
        prior_art_splits = split_passage_to_meet_model_token_limit(prior_art[1], partial(prompt_func, claim_element), model_name, min_n_splits=min_n_splits)
        n_passages_per_split = max(target_n_passages // len(prior_art_splits), 1)
        tasks.append(asyncio.create_task(
            agenerate_passages_for_prior_art_splits(prior_art_splits, partial(prompt_func, claim_element, n_passages_per_split=n_passages_per_split), model_name, claim_element_id=i, prior_art_id=prior_art[0], rate_limiter=rate_limiter)
        ))

    return tasks


def summarize_patent_prompt_func(serialized_patent_spec: str, instruction: str = None):
    if instruction is None:
        instruction = "Summarize the following patent specification in approximately 250 words. Your summary should not mention the 'patent', 'the specification', 'various embodiments', or any other language that suggests the summary is from a patent. You should not prefix the summary with 'Summary:' or any other prefatory language. The summary should be a standalone short paragraph."
    return [
        {"role": "system", "content": f"{instruction}"},
        {"role": "user", "content": f'Here is the specification to summarize: {serialized_patent_spec}'},
    ]


def summarize_summaries_prompt_func(summaries: List[str], instruction: Optional[str] = None):
    if instruction is None:
        instruction = "Summarize the following partial summaries of sequential portions of a patent specification in approximately 250 words. Your summary should not mention the 'patent', 'the specification', 'various embodiments', or any other language that suggests the summary is from a patent. You should not prefix the summary with 'Summary:' or any other prefatory language. The summary should be a standalone short paragraph"
    partial_summaries = '\n'.join(summaries)
    return [
        {"role": "system", "content": f"{instruction}"},
        {"role": "user", "content": f"Here are the sequential partial summaries I'd like you to summarize: {partial_summaries}"},
    ]


def short_summary_prompt_func(summary: str):
    return [
        {"role": "system", "content": "Condense the following summary to approximately 25 words. As an example, here is a long summary: 'The patent specification describes a closed-loop/semi-closed loop therapy modification system for insulin delivery in diabetes patients, focusing on managing over/under-delivery of medication. The system, incorporating a glucose sensor and insulin pump, triggers alarms based on blood glucose levels or insulin delivery discrepancies, prompting calibration of the sensor system and adjustments to therapy delivery parameters. The system aims to achieve and maintain target blood glucose levels effectively.\n\nThe invention relies on monitoring glucose levels continuously or periodically, adjusting insulin delivery based on the difference between measured and target values. Patient intervention may be required to approve calibration or therapy adjustments. The algorithm limits adjustments within preset boundaries, considering patient's basal pattern. It logs adjustments for patient reference and recommends new basal rates if adjustments consistently reach boundaries.\n\nSafeguards are incorporated to prevent excessive insulin delivery, including safety reviews, real-time calibration adjustment procedures, and safeguards when transitioning from closed-loop to open-loop systems. Additionally, the system accounts for insulin on board delays, ensuring insulin action is adequate before increasing doses. A model supervisory system and tolerance intervals further enhance safety and accuracy in therapy management.\n\nThe specification emphasizes versatility, adaptability, and patient-centric customization, aiming to optimize insulin therapy for diabetes patients. Various embodiments are discussed to address different scenarios and challenges that may arise during insulin delivery, ensuring patient safety and treatment effectiveness throughout the process.' Here is an initial attempt at a condensed summary: 'Description of a closed-loop/semi-closed loop system for insulin delivery in diabetes patients, focusing on managing medication delivery with glucose sensor and insulin pump for optimized therapy results.' Here is a second attempt at a condensed summary: 'Closed-loop/semi-closed loop therapy modification system for insulin delivery in diabetes patients, focusing on managing medication delivery with glucose sensor and insulin pump for optimized therapy results.' This attempt is better because it omits the 'Description of' preamble and focuses exclusively on describing the invention. Your condensed summary should not mention the 'patent', 'the specification', 'various embodiments', or any other language that suggests the summary is from a patent. You should not prefix the summary with 'Summary:' or any other prefatory language. The summary should be a standalone short paragraph or sentence."},
        {"role": "user", "content": f'Here is the long summary for you to summarize: {summary}'},
    ]


def topics_from_summary_prompt_func(summary: str):
    return [
        {"role": "system", "content": "Identify the main topics in the following summary. You may state up to 5 topics in a comma separated list. If you believe there are fewer than 5 topics, you may state fewer. If you believe there are more than 5 topics, you may state the most important 5. Each topic should be one exactly one word. As an example, here is a long summary: 'The patent specification describes a closed-loop/semi-closed loop therapy modification system for insulin delivery in diabetes patients, focusing on managing over/under-delivery of medication. The system, incorporating a glucose sensor and insulin pump, triggers alarms based on blood glucose levels or insulin delivery discrepancies, prompting calibration of the sensor system and adjustments to therapy delivery parameters. The system aims to achieve and maintain target blood glucose levels effectively.\n\nThe invention relies on monitoring glucose levels continuously or periodically, adjusting insulin delivery based on the difference between measured and target values. Patient intervention may be required to approve calibration or therapy adjustments. The algorithm limits adjustments within preset boundaries, considering patient's basal pattern. It logs adjustments for patient reference and recommends new basal rates if adjustments consistently reach boundaries.\n\nSafeguards are incorporated to prevent excessive insulin delivery, including safety reviews, real-time calibration adjustment procedures, and safeguards when transitioning from closed-loop to open-loop systems. Additionally, the system accounts for insulin on board delays, ensuring insulin action is adequate before increasing doses. A model supervisory system and tolerance intervals further enhance safety and accuracy in therapy management.\n\nThe specification emphasizes versatility, adaptability, and patient-centric customization, aiming to optimize insulin therapy for diabetes patients. Various embodiments are discussed to address different scenarios and challenges that may arise during insulin delivery, ensuring patient safety and treatment effectiveness throughout the process.' And this would be an acceptable output: 'Insulin delivery, Diabetes, Therapy system, Monitoring, Glucose sensor'. Topics should reference the specifics of the invention, not the patent or the specification. For example, 'Patent specification' would be an unacceptable topic. Do not include prefatory language such as 'Based on the provided summary, here are the main topics:' or 'The main topics are:', just return the five topics each separated by a comma."},
        {"role": "user", "content": f'Here is the long summary for you to choose topics from: {summary}'},
    ]


def parse_topics(topics: str) -> list[str]:
    topics_list = topics.split(',')
    topics_list = [t.strip(' \'\n') for t in topics_list]
    return topics_list


async def asummarize_patent(patent: parser.GoogleParsedPatent, rate_limiter: Optional[DistributedFixedWindowRateLimiter] = None, generator_config=None):
    if generator_config is None:
        generator_config = load_generator_config_from_env()
    serialized_patent_spec = patent.specification

    split_instruction = "Summarize the following portion of a patent specification in approximately 250 words:"
    spec_splits = split_passage_to_meet_model_token_limit(serialized_patent_spec, partial(summarize_patent_prompt_func, instruction=split_instruction), generator_config.model_name)

    if len(spec_splits) == 1:
        summarization = await make_get_completion_coro(
            generator_config.model_name, summarize_patent_prompt_func(serialized_patent_spec), rate_limiter=rate_limiter
        )
    else:
        tasks = []
        # TODO: hack
        for spec_split in spec_splits[:10]:
            tasks.append(
                make_get_completion_coro(
                    generator_config.model_name, summarize_patent_prompt_func(spec_split, instruction=split_instruction), rate_limiter=rate_limiter
                )
            )
        partial_summaries = await asyncio.gather(*tasks)
        if generator_config.model_name == 'claude-3-opus-20240229':
            partial_summaries = [post_process_summary_claude_opus(partial_summary) for partial_summary in partial_summaries]
        try:
            summarization = await make_get_completion_coro(
                generator_config.model_name, summarize_summaries_prompt_func(partial_summaries), rate_limiter=rate_limiter
            )
        except GenerationError as e:
            logger.error('Failed to summarize patent %s', patent.unique_id)
            summarization = ''

    if generator_config.model_name == 'claude-3-opus-20240229':
        summarization = post_process_summary_claude_opus(summarization)

    short_summary = await make_get_completion_coro(
        generator_config.model_name, short_summary_prompt_func(summarization), rate_limiter=rate_limiter
    )
    if generator_config.model_name == 'claude-3-opus-20240229':
        short_summary = post_process_summary_claude_opus(short_summary)

    topics = await make_get_completion_coro(
        generator_config.model_name, topics_from_summary_prompt_func(summarization), rate_limiter=rate_limiter
    )
    topics = parse_topics(topics)

    search_index.update_patent_by_id(patent, summary=summarization, short_summary=short_summary, topics=topics)

    return summarization


async def asummarize_patents(patents: list[parser.GoogleParsedPatent]):
    generator_config = load_generator_config_from_env()
    async with DistributedFixedWindowRateLimiter(token_rate_limit_by_model_name[generator_config.model_name], RATE_LIMIT_INTERVAL) as rate_limiter:
        tasks = [asummarize_patent(patent, rate_limiter=rate_limiter) for patent in patents]
        return await asyncio.gather(*tasks)


async def asummarize_patents_gpt4_turbo_preview(patents: list[parser.GoogleParsedPatent]):
    generator_config = load_generator_config_from_env()
    generator_config.model_name = 'gpt-4-turbo-preview'
    async with DistributedFixedWindowRateLimiter(token_rate_limit_by_model_name[generator_config.model_name], RATE_LIMIT_INTERVAL) as rate_limiter:
        tasks = [asummarize_patent(patent, rate_limiter=rate_limiter, generator_config=generator_config) for patent in patents]
        return await asyncio.gather(*tasks)


async def hydrate_prompt_with_patent(patent: parser.GoogleParsedPatent, generator_config: GeneratorConfig) -> partial[Callable[[str], list[dict]]]:
    claims = patent[1].claims
    all_patent_claim_elements = []
    for claim in claims:
        all_patent_claim_elements.extend(claim.claim_elements)
    serialized_claims = '\n'.join(all_patent_claim_elements)
    
    if generator_config.model_name == 'gpt-4':
        patent_summary = search_index.get_patent_summary(patent[1])
        if patent_summary is None:
            patent_summary = await asummarize_patent(patent[1])
        prompt_func = partial(prompt_summary_full_set_of_claims, patent_summary, serialized_claims)
        return prompt_func
    elif generator_config.model_name == 'claude-3-opus-20240229':
        return partial(prompt_full_spec_full_set_of_claims, patent[1].specification, serialized_claims)
    else:
        raise NotImplementedError(f"Model {generator_config} not supported.")


async def abulk_generate_passages(
        patent: tuple[int, parser.GoogleParsedPatent], 
        prior_art_sources: list[tuple[int, parser.GoogleParsedPatent]], claim_elements: list[tuple[int, str]],
        ranking_params: dict,
        ) -> AsyncIterator[GeneratedPassage]:
    # TODO: this function should be refactored to separate concerns
    generator_config = load_generator_config_from_env()
    async with DistributedFixedWindowRateLimiter(token_rate_limit_by_model_name[generator_config.model_name], RATE_LIMIT_INTERVAL) as rate_limiter:
        prompt_func = await hydrate_prompt_with_patent(patent, generator_config)
        tasks = []
        for prior_art_source_uid, prior_art_source in prior_art_sources:
            serialized_prior_art_source = prior_art_source.specification
            tasks.extend(
                create_generate_passages_tasks(
                    claim_elements, (prior_art_source_uid, serialized_prior_art_source), 
                    prompt_func,
                    generator_config.model_name,
                    target_n_passages=generator_config.target_n_passages,
                    min_n_splits=generator_config.min_n_splits,
                    rate_limiter=rate_limiter
            ))
        
        for coro in asyncio.as_completed(tasks):
            generated_passages = await coro
            # TODO: once we yield the whole list of generated passages for each awaited task, we can do the ranking step in the caller
            # Find matching claim element
            if generated_passages:
                claim_element_id = generated_passages[0].claim_element_id
                matching_claim_element = [c for c in claim_elements if c[0] == claim_element_id][0]
                if not all([p.claim_element_id == claim_element_id for p in generated_passages]):
                    logger.error('Claim element ids do not match')
                
                logger.info(f'Ranking {len(generated_passages)} passages')
                ranked_generated_passages = await ranking.arank_passages(generated_passages, matching_claim_element[1], **ranking_params)
                for generated_passage in ranked_generated_passages:
                    yield generated_passage


@lru_cache
def num_tokens_from_string(string: str, encoding_name: str = 'cl100k_base') -> int:
    """Returns the number of tokens in a text string."""
    encoding = tiktoken.get_encoding(encoding_name)
    num_tokens = len(encoding.encode(string))
    return num_tokens


@retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6), reraise=True)
def encode_openai(text_batch):
    embeddings = oai.embeddings.create(
        input=text_batch,
        model='text-embedding-3-large',
        dimensions=1024
    )
    return [e.embedding for e in embeddings.data]


def batch_encode_openai(texts):
    i = 0
    embeddings = []
    while i < len(texts):
        batch_tok_count = 0
        batch = []
        while i < len(texts) and batch_tok_count + num_tokens_from_string(texts[i]) < MAX_SEQ_LENGTH:
            batch_tok_count += num_tokens_from_string(texts[i])
            batch.append(texts[i])
            i += 1
        embeddings.extend(encode_openai(batch))
        
    return embeddings


import pathlib
import unittest
import asyncio
from functools import partial
from unittest.mock import patch, AsyncMock
from pprint import pprint


from patent_chart import parser
from patent_chart import generator
from patent_chart import utils
from patent_chart import filtering
from patent_chart import ranking

# NOTE: relic of when we were trying to parse the whole patent from the pdf.
# class TestParser(unittest.TestCase):
#     package_dir = pathlib.Path(__file__).parents[1]
#     patent_257_path = package_dir / 'test_data/US 6,484,257.pdf'
#     patent_131_path = package_dir / 'test_data/US 7,600,131.pdf'
#     patent_479_path = package_dir / 'test_data/US 5,870,479.pdf'
#     patent_448_path = package_dir / 'test_data/US7069448.pdf'
#     patent_449_path = package_dir / 'test_data/7069449.pdf'
#     patent_application_path = package_dir / 'test_data/US20210365512A1.pdf'
#     patent_325_path = package_dir / 'test_data/US7031325.pdf'
    
#     def test_parse_patent_from_pdf_path(self):
#         text_lines = parser.parse_text_lines_from_pdf_path(self.patent_257_path)
#         parsed_patent = parser.parse_patent_from_text_lines(text_lines)
#         # print(parser.group_parsed_patent_by_page(parsed_patent)[23].column_1)
#         # return
#         self.assertEqual(
#             parsed_patent.unique_id,
#             parser.PatentUniqueID(
#                 patent_number='6484257',
#                 country_code='US',
#                 kind_code='B1',
#             )
#         )
#         self.assertEqual(
#             parsed_patent.beginning_of_specification,
#             16
#         )
#         # claims = parser.parse_claims_from_parsed_patent(parsed_patent)
#         # for claim in claims.claims:
#         #     print(parser.serialize_claim(claim))
#         # return
#         # pprint(claims)
#         self.assertEqual(
#             parsed_patent.beginning_of_claims,
#             parser.BeginningOfClaims(
#                 page_index=23,
#                 col_index=0,
#                 line_index=16,
#             )
#         )
#         self.assertEqual(
#             parsed_patent.end_of_claims,
#             parser.EndOfClaims(
#                 page_index=23,
#                 col_index=1,
#                 line_index=74,
#             )
#         )

#         text_lines = parser.parse_text_lines_from_pdf_path(self.patent_131_path)
#         parsed_patent = parser.parse_patent_from_text_lines(text_lines)
#         self.assertEqual(
#             parsed_patent.unique_id,
#             parser.PatentUniqueID(
#                 patent_number='7600131',
#                 country_code='US',
#                 kind_code='B1',
#             )
#         )
#         self.assertEqual(
#             parsed_patent.beginning_of_specification,
#             12
#         )
#         self.assertEqual(
#             parsed_patent.beginning_of_claims,
#             parser.BeginningOfClaims(
#                 page_index=22,
#                 col_index=0,
#                 line_index=11,
#             )
#         )
#         self.assertEqual(
#             parsed_patent.end_of_claims,
#             parser.EndOfClaims(
#                 page_index=22,
#                 col_index=1,
#                 line_index=42,
#             )
#         )

#         text_lines = parser.parse_text_lines_from_pdf_path(self.patent_479_path)
#         parsed_patent = parser.parse_patent_from_text_lines(text_lines)
#         self.assertEqual(
#             parsed_patent.unique_id,
#             parser.PatentUniqueID(
#                 patent_number='5870479',
#                 country_code='US',
#                 kind_code='A',
#             )
#         )
#         self.assertEqual(
#             parsed_patent.beginning_of_specification,
#             4
#         )
#         self.assertEqual(
#             parsed_patent.beginning_of_claims,
#             parser.BeginningOfClaims(
#                 page_index=7,
#                 col_index=0,
#                 line_index=8,
#             )
#         )
#         self.assertEqual(
#             parsed_patent.end_of_claims,
#             parser.EndOfClaims(
#                 page_index=7,
#                 col_index=1,
#                 line_index=50,
#             )
#         )

#         text_lines = parser.parse_text_lines_from_pdf_path(self.patent_application_path)
#         parsed_patent = parser.parse_patent_from_text_lines(text_lines)
#         self.assertEqual(
#             parsed_patent.unique_id,
#             parser.PatentUniqueID(
#                 patent_number='20210365512',
#                 country_code='US',
#                 kind_code='A1',
#             )
#         )

#         self.assertEqual(
#             parsed_patent.beginning_of_specification,
#             15
#         )

#         # # TODO: doesn't work yet searching for specific prefatory language because there is none in this case. Would have to have some model to catch this one. Could count number of 'X.' bigrams on page, could use simple naive bayes model, could even just ask LLM.
#         # self.assertEqual(
#         #     parsed_patent.beginning_of_claims,
#         #     parser.BeginningOfClaims(
#         #         page_index=22,
#         #         col_index=0,
#         #         line_index=0,
#         #     )
#         # )
#         # # TODO: doesn't work for this one either
#         # self.assertEqual(
#         #     parsed_patent.end_of_claims,
#         #     parser.EndOfClaims(
#         #         page_index=24,
#         #         col_index=1,
#         #         line_index=0,
#         #     )
#         # )

#         parsed_patent = parser.parse_patent_from_pdf_path(self.patent_448_path)
#         self.assertEqual(
#             parsed_patent.unique_id,
#             parser.PatentUniqueID(
#                 patent_number='7069448',
#                 country_code='US',
#                 kind_code='B2',
#             )
#         )
#         self.assertEqual(
#             parsed_patent.beginning_of_specification,
#             4
#         )
#         self.assertEqual(
#             parsed_patent.beginning_of_claims,
#             parser.BeginningOfClaims(
#                 page_index=6,
#                 col_index=1,
#                 line_index=25,
#             )
#         )
#         self.assertEqual(
#             parsed_patent.end_of_claims,
#             parser.EndOfClaims(
#                 page_index=7,
#                 col_index=1,
#                 line_index=23,
#             )
#         )

#         # TODO: doesnt work because each page of 449 is a figure. None of the text is selectable. Each page contains a pdfminer.six LTFigure object, which might be an embedded pdf. see comment in pdfminer.six/pdfminer/layout.py: class LTFigure(LTLayoutContainer):
#         """Represents an area used by PDF Form objects.

#         PDF Forms can be used to present figures or pictures by embedding yet
#         another PDF document within a page. Note that LTFigure objects can appear
#         recursively.
#         """
#         # So we might just need to recurse through LTFigure objects when we encounter them as the page contents.
#         # parsed_patent = parser.parse_patent_from_pdf_path(self.patent_449_path)

#         # TODO: test specific expected lines

#     def test_parse_some_more_patents(self):
#         parsed_patent = parser.parse_patent_from_pdf_path(self.patent_325_path)
#         self.assertEqual(
#             parsed_patent.unique_id,
#             parser.PatentUniqueID(
#                 patent_number='7031325',
#                 country_code='US',
#                 kind_code='B1',
#             )
#         )
#         self.assertEqual(
#             parsed_patent.beginning_of_specification,
#             8
#         )

#         self.assertEqual(
#             parsed_patent.beginning_of_claims,
#             parser.BeginningOfClaims(
#                 page_index=13,
#                 col_index=1,
#                 line_index=50,
#             )
#         )
        
#         self.assertEqual(
#             parsed_patent.end_of_claims,
#             parser.EndOfClaims(
#                 page_index=15,
#                 col_index=1,
#                 line_index=43,
#             )
#         )

#     # def test_parse_broad_set_us_patents(self):
#     #     us_patents_dir = self.package_dir / 'us_patents_1980-2020'
#     #     path_to_expected = {
#     #         'US6727522.pdf': {
#     #             'unique_id': parser.PatentUniqueID(
#     #                 patent_number='6727522',
#     #                 country_code='US',
#     #                 kind_code='B2',
#     #             ),
#     #             'beginning_of_specification': 10,
#     #             'beginning_of_claims': parser.BeginningOfClaims(
#     #                 page_index=6,
#     #                 col_index=1,
#     #                 line_index=25,
#     #             ),
#     #             'end_of_claims': parser.EndOfClaims(
#     #                 page_index=7,
#     #                 col_index=1,
#     #                 line_index=23,
#     #             ),
#     #             'randomly_selected_claim_element': (
#     #                 ()
#     #             )
#     #         },
#     #     }
        
#     #     for pdf_path in us_patents_dir.glob('*.pdf'):
#     #         print(pdf_path.name)
#     #         if pdf_path.name == 'USRE48787.pdf':
#     #             continue
#     #         # if pdf_path.name == 'US6727522.pdf':
#     #         #     continue
#     #         # parsed_patent = parser.parse_patent_from_pdf_path(pdf_path)
#     #         # print(parsed_patent.unique_id)

#     #         # claims = parser.parse_claims_from_parsed_patent(parsed_patent)
#     #         unique_id = parser.parse_patent_unique_id_from_pdf_path(pdf_path)
#     #         print(unique_id)

#     def test_parse_claims_from_parsed_patent(self):
#         parsed_patent = parser.parse_patent_from_pdf_path(self.patent_257_path)
#         claims = parser.parse_claims_from_parsed_patent(parsed_patent)
#         first_claim = parser.serialize_claim(claims.claims[0])
#         self.assertEqual(
#             first_claim,
#             "1. A software architecture for conducting a plurality of 15cryptographic sessions over a distributed computingenvironment, comprising:a registration entity or registry residing within a mamserver entity;an agent server entity communicating with said mam 20server;a client entity communicating with said main server andagent server;a plurality of distributed networked computers providinga mechanism for executing said main server entity,agent server entity, and client entity;a defined protocol for initiating secure communicationbetween the main server and agent server; over saidnetwork; anda system for providing one or more communicationsessions among the main server, agent server and cliententity for implementing a client decrypted bandwidthreconstitution which enables the recombination of individual parts of the decrypted client bandwidth among Nagents processing in parallel."
#         )

#         claim_elements = parser.serialize_claim_elements(claims.claims[0])
#         # TODO: see 'over said network;' parsed as it's own element. Apparently can't rely on ';' separating claim elements in every case.
#         self.assertEqual(
#             claim_elements,
#             ['1. A software architecture for conducting a plurality of 15cryptographic sessions over a distributed computingenvironment, comprising:', 'a registration entity or registry residing within a mamserver entity;', 'an agent server entity communicating with said mam 20server;', 'a client entity communicating with said main server andagent server;', 'a plurality of distributed networked computers providinga mechanism for executing said main server entity,agent server entity, and client entity;', 'a defined protocol for initiating secure communicationbetween the main server and agent server;', 'over saidnetwork;', 'anda system for providing one or more communicationsessions among the main server, agent server and cliententity for implementing a client decrypted bandwidthreconstitution which enables the recombination of individual parts of the decrypted client bandwidth among Nagents processing in parallel.']
#         )

#         parsed_patent = parser.parse_patent_from_pdf_path(self.patent_448_path)
#         claims = parser.parse_claims_from_parsed_patent(parsed_patent)
        
#         serialized_claim_1 = parser.serialize_claim_elements(claims.claims[0])
#         self.assertEqual(
#             serialized_claim_1,
#             ['1. A system for cryptographic processing of input data ona parallel processor array that includes a plurality of processors, comprising:', 'a format filter adapted to extract control data and maindata from the input data;', 'a control unit adapted to receive the control data from saidformat filter, and to forward, based at least in part on thecontrol data, at least one respective control parameterand at least one respective cryptographic parameter toeach of the plurality of processors;', 'a first distributor adapted to receive the main data fromsaid format filter, and to distribute to each of theplurality of processors a respective at least a portion ofthe main data;', 'a second distributor adapted to receive respective outputinformation from each of the plurality of processors,and to generate, based at least in part on the respectiveoutput information, output data;', 'wherein each of the plurality of processors is adapted togenerate its respective output information based at leastin part on the control parameters and the cryptographicparameters, and the output data is a cryptographicprocessing result.']
#         )

#         parsed_patent = parser.parse_patent_from_pdf_path(self.patent_479_path)
#         claims = parser.parse_claims_from_parsed_patent(parsed_patent)
#         claim_elements = parser.serialize_claim_elements(claims.claims[-1])
#         self.assertEqual(
#             claim_elements,
#             # TODO: need to fix these floating line number lines that get tacked on (3016 should be 16)
#             ['3016. A device for cryptographically processing datapackets, each of the data packets belonging to at least one ofa plurality of channels, the device comprising:', 'identification means for identifying the at least one channel to which a data packet belongs;', 'processing means for cryptographically processing thedata packet, wherein the processing means include afirst processing unit and a second processing unit;', 'memory means for storing information, associated witheach of the plurality of channels, for processing datapackets from each of the plurality a channels;', 'andcontrol means for selecting information associated withthe at least one channel which the data packet wasidentified as belonging to, wherein the control meansare designed to assign, on the basis of the identificationof the data packet, the data packet to one of the first andsecond processing units and to process the data packetwith the aid of the selected information.']
#         )

#     def test_serialize_specification_from_parsed_patent(self):
#         parsed_patent = parser.parse_patent_from_pdf_path(self.patent_257_path)
#         specification = parser.serialize_specification_from_parsed_patent(parsed_patent)

#         self.assertEqual(
#             specification[:13],
#             ['1', 'SYSTEM AND METHOD FOR MAINTAINING', 'N NUMBER OF SIMULTANEOUS', 'CRYPTOGRAPHIC SESSIONS USING A', 'DISTRIBUTED COMPUTING', 'ENVIRONMENT', 'FIELD OF THE INVENTION', 'The field of the present invention relates generally to the', 'encryption and decryption of data conducted over a distrib', 'uted computer network. In particular, the field of the inven', 'tion relates to a software architecture for conducting a', 'plurality of cryptographic sessions managed over a distrib', 'uted computing environment.']
#         )

#         parsed_patent = parser.parse_patent_from_pdf_path(self.patent_131_path)
#         specification = parser.serialize_specification_from_parsed_patent(parsed_patent)
#         self.assertEqual(
#             specification[-1],
#             'claims and their full scope of equivalents.'
#         )


class TestParseGooglePatents(unittest.TestCase):
    package_dir = pathlib.Path(__file__).parents[1]

    def test_parse_broad_set_us_patents_google(self):
        n_parse_failures = 0
        n_parse_claims_failures = 0
        us_patents_dir = self.package_dir / 'us_patents_1980-2020'
        for pdf_path in us_patents_dir.glob('*.pdf'):
            parsed_google_patent = parser.parse_google_patent_from_pdf_path(pdf_path)
            if parsed_google_patent is None:
                print(f'Failed to parse {pdf_path}')
                n_parse_failures += 1
            else:
                claims = parser.parse_claims_from_google_parsed_patent(parsed_google_patent)
                if claims is None:
                    print(f'Failed to parse claims from {pdf_path}')
                    n_parse_claims_failures += 1

        self.assertEqual(n_parse_failures, 0)
        self.assertEqual(n_parse_claims_failures, 0)


class TestFiltering(unittest.TestCase):
    def test_filtering(self):
        generated_passages = [
            generator.GeneratedPassage(
                claim_element_id=0,
                prior_art_source_id=1,
                text='generated passage',
                model_id='model',
            ),
            generator.GeneratedPassage(
                claim_element_id=0,
                prior_art_source_id=2,
                text='another generated passage',
                model_id='model',
            ),
        ]
        serialized_prior_art_source = 'generated passage'
        filter_pipeline = [partial(filtering.inclusion, super_string=serialized_prior_art_source), partial(filtering.min_length, min_length=5)]

        filter_mask = filtering.get_filter_mask([p.text for p in generated_passages], *filter_pipeline)
        self.assertEqual(len(filter_mask), 2)
        self.assertEqual(
            filter_mask,
            [True, False]
        )

        filtered_passages = filtering.apply_filter_pipeline([p.text for p in generated_passages], *filter_pipeline)
        self.assertEqual(len(filtered_passages), 1)
        self.assertEqual(
            filtered_passages,
            ['generated passage']
        )


class TestAsyncGenerator(unittest.IsolatedAsyncioTestCase):
    package_dir = pathlib.Path(__file__).parents[1]
    patent_448_path = package_dir / 'test_data/US7069448.pdf'
    patent_131_path = package_dir / 'test_data/US 7,600,131.pdf'
    patent_257_path = package_dir / 'test_data/US 6,484,257.pdf'

    @patch('patent_chart.generator.aopenai_chat_completion_request_with_retry', new_callable=AsyncMock)
    async def test_async_generate(self, mock_aopenai_chat_completion_request_with_retry):
        """
        Does not test parsing of openai output. Generated passages will typically be multiple passages separated by newlines rather than just a single passage 'generated_passage'
        """
        mock_aopenai_chat_completion_request_with_retry.return_value = ({
            'choices': [
                {
                    'message': {
                        'content': 'generated_passage'
                    },
                }
            ]
        })
        parsed_patent_448 = parser.parse_google_patent_from_pdf_path(self.patent_448_path)
        parsed_patent_131 = parser.parse_google_patent_from_pdf_path(self.patent_131_path)
        parsed_patent_257 = parser.parse_google_patent_from_pdf_path(self.patent_257_path)

        claims_448 = parsed_patent_448.claims
        
        serialized_claim_elements = []
        for claim in claims_448:
            serialized_claim_elements.extend(claim.claim_elements)

        generated_passages = set()
        full_generated_passages = []
        async for generated_passage in generator.abulk_generate_passages(
            (1, parsed_patent_448),
            [(1, parsed_patent_131), (2, parsed_patent_257)],
            [(i, serialized_claim_elements[i]) for i in range(len(serialized_claim_elements))],
            {'ranking_model_version': 'text-embedding-3-large'}
        ):
            generated_passages.add(
                (generated_passage.claim_element_id, generated_passage.prior_art_source_id)
            )
            full_generated_passages.append(generated_passage)
        
        self.assertEqual(
            len(generated_passages),
            2 * len(serialized_claim_elements)
        )

        self.assertEqual(
            generated_passages,
            set(
                [
                    (j, i) for i in range(1, 3) for j in range(len(serialized_claim_elements))
                ]
            )
        )

    def test_parse_gpt4_responses(self):
        """
        These passages were extracted from patent 131
        """
        case1 = "\"Many methods to perform cryptography are well known in\nthe art and are discussed, for example, in Applied Cryptogra\nphy, Bruce Schneier, John Wiley & Sons, Inc. (1996, rd\nEdition), herein incorporated by reference. In order to\nimprove the speed of cryptography processing, specialized\ncryptography accelerator chips have been developed.\"\n\n\"Moreover, the architecture of prior art chips does not allow\nfor the processing of cryptographic data at rates sustainable\nby the network infrastructure in connection with which these\nchips are generally implemented. This can result in noticeable 60\ndelays when cryptographic functions are invoked, for\nexample, in e-commerce transactions.\"\n\n\"In one aspect, the present invention provides a cryptogra-\nphy acceleration chip. The chip includes a plurality of cryp\ntography processing engines, and a packet distributor unit.\""

        parsed_case1 = generator.post_process_selected_passage_gpt4(case1)

        self.assertEqual(
            parsed_case1,
            [
                'Many methods to perform cryptography are well known in\nthe art and are discussed, for example, in Applied Cryptogra\nphy, Bruce Schneier, John Wiley & Sons, Inc. (1996, rd\nEdition), herein incorporated by reference. In order to\nimprove the speed of cryptography processing, specialized\ncryptography accelerator chips have been developed.',
                'Moreover, the architecture of prior art chips does not allow\nfor the processing of cryptographic data at rates sustainable\nby the network infrastructure in connection with which these\nchips are generally implemented. This can result in noticeable 60\ndelays when cryptographic functions are invoked, for\nexample, in e-commerce transactions.',
                'In one aspect, the present invention provides a cryptogra-\nphy acceleration chip. The chip includes a plurality of cryp\ntography processing engines, and a packet distributor unit.'
            ]
        )

        case2 = "\"Full support for IPSec Security Association Database lookup, including wildcard rules, overlapping rules, and complete ordering of database entries.\"\n\"Because of the pipelined design, throughput is gated by the slowest set of stages.\"\n\"The micro-engine is started by an event-driven mechanism.\""

        parsed_case2 = generator.post_process_selected_passage_gpt4(case2)

        self.assertEqual(
            parsed_case2,
            [
                'Full support for IPSec Security Association Database lookup, including wildcard rules, overlapping rules, and complete ordering of database entries.',
                'Because of the pipelined design, throughput is gated by the slowest set of stages.',
                'The micro-engine is started by an event-driven mechanism.'
            ]
        )

        case3 = "1. \"For every new packet, the distributor com\npletes the sequential portions of IPSec processing, and\nassigns the packet to the next free engine. Once the engine\ncompletes processing the packet, the processed packet is\nplaced in a retirement buffer.\"\n\n2. \"The distributor unit with an order maintenance packet\nretirement unit.\"\n\n3. \"The packet distributor unit 306 then distributes the security\nassociation information (SA) received from the packet classifier\namong a plurality of cryptography processing engines 316, on the chip 200, for security processing.\""

        parsed_case3 = generator.post_process_selected_passage_gpt4(case3)

        self.assertEqual(
            parsed_case3,
            [
                'For every new packet, the distributor com\npletes the sequential portions of IPSec processing, and\nassigns the packet to the next free engine. Once the engine\ncompletes processing the packet, the processed packet is\nplaced in a retirement buffer.',
                'The distributor unit with an order maintenance packet\nretirement unit.',
                'The packet distributor unit 306 then distributes the security\nassociation information (SA) received from the packet classifier\namong a plurality of cryptography processing engines 316, on the chip 200, for security processing.'
            ]
        )

        case4 = "\"When an outbound packet is received by the packet classifier on a cryptography acceleration chip in accordance with the present invention, its header is parsed ( 652) and a SPD lookup is performed (654).\"\n\n\"Once in the system, a SAD lookup is conducted (660).\"\n\n\"The input to ACE consists of packet classification fields: src/dst address, src/dst ports, and protocol. The output of ACE is an IPSec Security Association matching entry, if one exists, for this classification information within the IPSec Security Association Database.\""

        parsed_case4 = generator.post_process_selected_passage_gpt4(case4)

        self.assertEqual(
            parsed_case4,
            [
                'When an outbound packet is received by the packet classifier on a cryptography acceleration chip in accordance with the present invention, its header is parsed ( 652) and a SPD lookup is performed (654).',
                'Once in the system, a SAD lookup is conducted (660).',
                'The input to ACE consists of packet classification fields: src/dst address, src/dst ports, and protocol. The output of ACE is an IPSec Security Association matching entry, if one exists, for this classification information within the IPSec Security Association Database.'
            ]
        )

    # NOTE: relic of when we were trying to parse the whole patent from the pdf.
    # def test_cite_passages(self):
    #     test_passages = [
    #         'Many methods to perform cryptography are well known in\nthe art and are discussed, for example, in Applied Cryptogra\nphy, Bruce Schneier, John Wiley & Sons, Inc. (1996, rd\nEdition), herein incorporated by reference. In order to\nimprove the speed of cryptography processing, specialized\ncryptography accelerator chips have been developed.',
    #         'Moreover, the architecture of prior art chips does not allow\nfor the processing of cryptographic data at rates sustainable\nby the network infrastructure in connection with which these\nchips are generally implemented. This can result in noticeable 60\ndelays when cryptographic functions are invoked, for\nexample, in e-commerce transactions.',
    #         'In one aspect, the present invention provides a cryptogra-\nphy acceleration chip. The chip includes a plurality of cryp\ntography processing engines, and a packet distributor unit.'
    #     ]

    #     test_passages_with_newlines_removed = [
    #         'Many methods to perform cryptography are well known in the art and are discussed, for example, in Applied Cryptogra phy, Bruce Schneier, John Wiley & Sons, Inc. (1996, rd Edition), herein incorporated by reference. In order to improve the speed of cryptography processing, specialized cryptography accelerator chips have been developed.',
    #         'The packet distributor unit is configured to receive data packets and matching classification information for the packets;',
    #         'The method involves receiving data packets on a cryptography acceleration chip, processing the data packets and matching classification information for the packets,',
    #         'Most functions of the distributor are performed via dedicated hardware assist logic as opposed to microcode, since the distributor 206 is directly in the critical path of per-packet processing.',
    #         'When an outbound packet is received by the packet classifier on a cryptography acceleration chip in accordance with the present invention, its header is parsed ( 652) and a SPD lookup is performed (654).',
    #         'Distributor Microcode Overview In one implementation of the present invention, the distributor unit has a micro-engine large register file (128 entries by 32-bits), good microcode RAM size (128 entries by 96-bits), and a simple three stage pipeline design that is visible to the instruction set via register read delay slots and loaded from the system port at power-up time, and is authenticated in order to achieve PIPS 140-1 compliance.',
    #         'The retirement unit then extracts processed packets out of the retirement buffer in the same order that the chip originally received the packets, and outputs the processed packets.',
    #         'packet header information is sent to a packet classifier unit 204 where a classification engine rapidly determines security association information required for processing the packet,',
    #         'Strong ordering may be maintained in a number of ways, for example, by assigning a new packet to the next free cryptography processing unit in strict round-robin sequence. Packets are retired in the same sequence as units complete processing, thus ensuring order maintenance.',
    #         # 'Inbound packets ... if sequence check enabled for inbound, check & update sequence mask; update Engine scheduling status; mark packet descriptor as free; add back to free pool; II Schedule write.', # NOTE this fails because there is a chunk missing from the middle of the passage
    #         'Once in the system, a SAD lookup is conducted (660). Ifno matching SAD entry is found (662) one is created (664) in the IPSec Security Association Database. The packet is encapsulated (666), encrypted and authenticated (668). The encrypted packet is then sent out of the system (670).',
    #     ]

    #     parsed_patent_131 = parser.parse_patent_from_pdf_path(self.patent_131_path)

    #     citations = []
    #     for passage in test_passages + test_passages_with_newlines_removed:
    #         # Will raise error if citation not found
    #         citations.append(utils.cite_passage(passage, parsed_patent_131))

    #     # "Many methods..." passage should have same citation line pair whether cited with newlines in or not
    #     self.assertEqual(citations[0], citations[3])

    #     self.assertEqual(
    #         [c[0].page_num for c in citations],
    #         [
    #             12,
    #             12,
    #             12,
    #             12,
    #             12,
    #             12,
    #             15,
    #             18,
    #             18,
    #             15,
    #             14,
    #             15,
    #             # 0,,
    #             18
    #         ]
    #     )

    #     for citation in citations:
    #         print(citation)
    #         print()


import datetime
import logging

from pinecone.grpc import PineconeGRPC as Pinecone
from pinecone import PineconeException
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
    retry_if_not_exception_message,
    retry_if_not_exception_type
) 

logger = logging.getLogger(__name__) 
pc = Pinecone()

class IndexError(Exception):
    pass

spec_index = pc.Index("patents-spec-embedding")
claims_index = pc.Index("patents-claim-embedding")


def doc_with_prefix_exists(index, prefix):
    try:
        next(index.list(prefix=prefix))
        return True
    except StopIteration:
        return False


@retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6), reraise=True, retry=retry_if_not_exception_type(IndexError))
def index_spec_embeddings(patent_doc_id:str, filing_date:str, priority_date:str, embeddings:list[list[float]], positions:list[tuple[int]], retry=retry_if_not_exception_type(PineconeException)):
    if filing_date:
        filing_unix_timestamp = datetime.datetime.strptime(filing_date, "%Y-%m-%d").timestamp()
    if priority_date:
        priority_unix_timestamp = datetime.datetime.strptime(priority_date, "%Y-%m-%d").timestamp()

    
    vectors = []
    for i, embedding in enumerate(embeddings):
        metadata = {
            "patent_doc_id": patent_doc_id, 
            "start_pos": positions[i][0], 
            "end_pos": positions[i][1]
        }
        if filing_date:
            metadata["filing_unix_timestamp"] = filing_unix_timestamp
        if priority_date:
            metadata["priority_unix_timestamp"] = priority_unix_timestamp
        
        vectors.append({
            "id": f"{patent_doc_id}-{i}", 
            "values": embedding,
            "metadata": metadata,
        })

    if not vectors:
        raise IndexError("No spec embeddings to index")

    spec_index.upsert(vectors)


@retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6), reraise=True, retry=retry_if_not_exception_type(IndexError))
def index_claim_embeddings(patent_doc_id:str, filing_date:str, priority_date:str, embeddings:list[list[float]]):
    if filing_date:
        filing_unix_timestamp = datetime.datetime.strptime(filing_date, "%Y-%m-%d").timestamp()
    if priority_date:
        priority_unix_timestamp = datetime.datetime.strptime(priority_date, "%Y-%m-%d").timestamp()

    vectors = []
    for i, embedding in enumerate(embeddings, start=1):
        metadata = {
            "patent_doc_id": patent_doc_id, 
            "claim_id": i
        }
        if filing_date:
            metadata["filing_unix_timestamp"] = filing_unix_timestamp
        if priority_date:
            metadata["priority_unix_timestamp"] = priority_unix_timestamp
        
        vectors.append({
            "id": f"{patent_doc_id}-{i-1}", 
            "values": embedding,
            "metadata": metadata,
        })

    if not vectors:
        raise IndexError("No claim embeddings to index")

    claims_index.upsert(vectors)


@retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6), reraise=True)
def query_spec_embedding(query: list[float], top_k: int, filter: dict | None = None, **kwargs):
    return spec_index.query(query, top_k=top_k, filter=filter, **kwargs)['matches']


@retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6), reraise=True)
def get_claim_embeddings_by_patent_doc_id(patent_doc_id:str):
    ids = list(claims_index.list(prefix=patent_doc_id))
    if not ids:
        return []
    ids = ids[0]
    records = claims_index.fetch(ids)
    return records['vectors']


def get_claim_embeddings_values_by_patent_doc_id(patent_doc_id:str):
    values = []
    records = get_claim_embeddings_by_patent_doc_id(patent_doc_id)
    for record in records:
        record_values = records[record]['values']
        values.append(record_values)
    return values


@retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6), reraise=True)
def get_spec_embeddings_by_patent_doc_id(patent_doc_id:str):
    ids = list(spec_index.list(prefix=patent_doc_id))
    if not ids:
        return []
    ids = ids[0]
    records = spec_index.fetch(ids)
    return records['vectors']


def get_spec_embeddings_values_by_patent_doc_id(patent_doc_id:str):
    values = []
    records = get_spec_embeddings_by_patent_doc_id(patent_doc_id)
    for record in records:
        record_values = records[record]['values']
        values.append(record_values)
    return values



import requests
import pathlib
import time
import typing
import re
import xml.etree.ElementTree as ET
from datetime import datetime
from functools import cached_property
from logging import getLogger

from langchain.text_splitter import RecursiveCharacterTextSplitter
from bs4 import BeautifulSoup

from patent_chart.data_structures import GoogleParsedPatent, PatentUniqueID, GoogleClaim

logger = getLogger(__name__)


class GoogleParseError(Exception):
    pass


def google_request_with_retry(url, max_retries=3):
    for i in range(1, max_retries + 1):
        response = requests.get(url)
        if response.status_code == 200:
            return response
        else:
            logger.warning('Failed to request %s. Response code: %s. Error: %s', url, response.status_code, response.text)
            logger.warning('Retrying %s', url)
            time.sleep(i**2)
    return None

def parse_google_claims(claims:str) -> list[GoogleClaim]:
    # Trim beginning of claims
    claim_begin = claims.find('1.')

    trimmed_claims = claims[claim_begin:]
    claim_lines = [l.strip() for l in trimmed_claims.split('\n') if l.strip()]

    claim_number = 1
    google_claims = []
    claim = GoogleClaim(
        claim_number=claim_number,
        claim_elements=[]
    )
    for line in claim_lines:
        if line.startswith(str(claim_number + 1)):
            claim_number += 1
            google_claims.append(claim)
            claim = GoogleClaim(
                claim_number=claim_number,
                claim_elements=[]
            )
        claim.claim_elements.append(line)

    return google_claims


filing_date_regex = re.compile(r'filing|filed')


class GooglePatent:
    def __init__(self, patent_id, html_text=None) -> None:
        self.patent_id = patent_id
        if html_text is None:
            self.soup = BeautifulSoup(requests.get(f"https://us-west1-careful-ai.cloudfunctions.net/google-patents?patent_id={patent_id}").text, 'html.parser')
        else:
            self.soup = BeautifulSoup(html_text, 'html.parser')
    
    @classmethod
    def from_html_text(cls, patent_id, html_text):
        return cls(patent_id, html_text)

    def _get_index_of_priority_date(self, table) -> int:
        if table:
            for index, span in enumerate(table.find("div", class_="thead").find_all("span", class_="th")):
                if span.get_text().strip().lower() == "priority date":
                    return index
        return -1

    def _extract_priority_dates(self, table):
        dates = []
        if table:
            for row in table.find("div", class_="tbody").find_all("div", class_="tr"):
                date_string = row.find_all("span")[self._get_index_of_priority_date(table)].get_text().strip()
                dates.append(datetime.strptime(date_string, "%Y-%m-%d"))
        return dates

    def _extract_priority_dates_from_parent_apps(self):
        if self.soup.find("h3", id="parentApplications"):
            if self.soup.find("h3", id="parentApplications").find_next_sibling():
                table = self.soup.find("h3", id="parentApplications").find_next_sibling().find("div", class_="table")
                return self._extract_priority_dates(table)
        return []
    
    def _extract_priority_dates_from_child_apps(self):
        if self.soup.find("h3", id="applicationChildApps"):
            if self.soup.find("h3", id="applicationPriorityApps").find_next_sibling():
                table = self.soup.find("h3", id="applicationPriorityApps").find_next_sibling().find("div", class_="table")
            return self._extract_priority_dates(table)
        return []

    def _extract_priority_dates_from_priority_apps(self):
        if self.soup.find("h3", id="applicationPriorityApps"):
            if self.soup.find("h3", id="applicationPriorityApps").find_next_sibling():
                table = self.soup.find("h3", id="applicationPriorityApps").find_next_sibling().find("div", class_="table")
            return self._extract_priority_dates(table)
        return []
    
    @cached_property
    def filing_date(self) -> datetime:
        application_timeline = self.soup.find('application-timeline')
        if not application_timeline:
            return None
        filing_date = application_timeline.find('div', class_=filing_date_regex)
        if filing_date:
            return datetime.strptime(filing_date.text, "%Y-%m-%d")
        return None

    @cached_property
    def inferred_priority_date(self):
        # TODO: figure out validation w/ prior art link
        # TODO: add child patents
        dates = self._extract_priority_dates_from_parent_apps() + self._extract_priority_dates_from_child_apps() + self._extract_priority_dates_from_priority_apps()
        if dates:
            # TODO: validate this logic - maybe take earliest one
            return min(dates)
        return None

    @cached_property
    def pdf_url(self) -> str:
        for a in self.soup.find_all("a"):
            if a.get_text().strip() == "Download PDF":
                return a.attrs["href"]
        return None

    @cached_property
    def prior_art_url(self) -> str:
        for a in self.soup.find_all("a"):
            if a.get_text().strip() == "Find Prior Art":
                if a.attrs["href"] != "#":
                    return a.attrs["href"]
        return None
    
    @cached_property
    def abstract(self) -> str:
        return self.soup.find("div", class_="abstract").get_text().strip() if self.soup.find("div", class_="abstract") else None

    @cached_property
    def title(self) -> str:
        return self.soup.find("h1", id="title").get_text().strip() if self.soup.find("h1", id="title") else None

    @cached_property
    def specification(self) -> str:
        return str(self.soup.find('section', id='description')) if self.soup.find('section', id='description') else None

    @cached_property
    def specification_text(self) -> str:
        specification = self.soup.find('section', id='description')
        return specification.get_text() if specification else None

    @cached_property
    def claims(self) -> str:
        return str(self.soup.find('section', id='claims')) if self.soup.find('section', id='claims') else None

    @cached_property
    def claims_text(self) -> str:
        claims = self.soup.find('section', id='claims')
        return claims.get_text() if claims else None

    @cached_property
    def jurisdiction(self) -> str:
        if self.soup.find("section", class_="knowledge-card"):
            if self.soup.find("section", class_="knowledge-card").find("header"):
                if self.soup.find("section", class_="knowledge-card").find("header").find("p"):
                    return self.soup.find("section", class_="knowledge-card").find("header").find("p").get_text().strip()
        return None
    
    @cached_property
    def known_references(self) -> list[str]:
        header = self.soup.find('h3', id='patentCitations')
        table = header.find_next_sibling()
        return [a.text for a in table.find_all('a')]


def parse_google_patent(patent_unique_id: PatentUniqueID) ->  GoogleParsedPatent | None:
    # Uses selenium to get dates -- probably faster way but whatever
    patent = GooglePatent(patent_unique_id)

    return GoogleParsedPatent(
        unique_id=patent_unique_id,
        specification=patent.specification_text,
        claims=patent.claims_text,
        title=patent.title,
        text=str(patent.soup),
        text_format='google_html_with_selenium',
        filing_date=patent.filing_date.strftime('%Y-%m-%d'),
        priority_date=patent.inferred_priority_date.strftime('%Y-%m-%d'),
    )


def extract_text_recursive(element: ET.Element):
    text = ''
    if element.text:
        text += element.text
    for child in element:
        text += extract_text_recursive(child)
    if element.tail:
        text += element.tail
    return text


def chunk_written_description_with_pos(description: ET.Element, chunk_size_words=8191, chunk_overlap_words=1024) -> tuple[list[str], list[tuple[int]]]:
    text = ''
    for elem in description:
        if elem.tag not in ['p', 'heading']:
            elem_text = ' '.join([extract_text_recursive(child) for child in elem])
        else:
            elem_text = extract_text_recursive(elem)
        text += elem_text
        if elem.tag == 'p':
            text += '\n'

    return chunk_written_description_from_text_with_pos(text, chunk_size_words, chunk_overlap_words)


def chunk_written_description_from_text_with_pos(text: str, chunk_size_words=8191, chunk_overlap_words=1024) -> tuple[list[str], list[tuple[int]]]:
    text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(
        model_name='gpt-4',
        chunk_size=chunk_size_words,
        chunk_overlap=chunk_overlap_words,
    )

    chunks = text_splitter.split_text(text)

    # Find start and end char pos for each chunk
    pos = 0
    starts = []
    ends = []
    for chunk in chunks:
        starts.append(pos)
        pos += len(chunk)
        ends.append(pos)

    return chunks, list(zip(starts, ends))


def extract_chunked_claims_and_spec(patent: GoogleParsedPatent):
    claims = patent.claims
    specification = patent.specification
    google_claims = parse_google_claims(claims.strip(' \n'))
    claims = ['\n'.join(claim.claim_elements) for claim in google_claims]
    specification = specification.strip(' \n')
    specification_chunks_with_pos = chunk_written_description_from_text_with_pos(specification)
    return claims, specification_chunks_with_pos


# Deprecated, but keeping for reference
# def parse_google_patent(patent_unique_id: PatentUniqueID) ->  GoogleParsedPatent | None:
#     url = f"https://patents.google.com/patent/{patent_unique_id}/en"

#     logger.info('Requesting %s', url)

#     # Send an HTTP GET request to fetch the webpage
#     response = google_request_with_retry(url)

#     # Parse the HTML content of the page using BeautifulSoup
#     soup = BeautifulSoup(response.text, 'html.parser')

#     specification = soup.find('section', {'itemprop': 'description'})
#     claims = soup.find('section', {'itemprop': 'claims'})
    

#     if specification is None or claims is None:
#         logger.warning('Failed to parse %s', url)
#         return None

#     title = soup.find('td', {'itemprop': 'title'})
    
#     title_text = None
#     if title is not None:
#         title_text = title.text.strip(' \n')

#     text = str(soup)
#     text_format = 'google_html'

#     return GoogleParsedPatent(
#         unique_id=patent_unique_id,
#         title=title_text,
#         text=text,
#         text_format=text_format,
#     )

# if __name__ == '__main__':
#     package_dir = pathlib.Path(__file__).parents[1]
#     us_patents_dir = package_dir / 'us_patents_1980-2020'
#     for pdf_path in us_patents_dir.glob('*.pdf'):
#         unique_id = parser.parse_patent_unique_id_from_pdf_path(pdf_path)
#         if unique_id is not None:
#             google_unique_id = f'{unique_id.country_code}{unique_id.patent_number}{unique_id.kind_code or ""}'
#             result = parse_google_patent(google_unique_id)
#             if result is not None:
#                 specification, claims = result
#                 assert specification != ''
#                 assert claims != ''
#             else:
#                 print(f'Failed to parse {unique_id}')
#         else:
#             print(f'Failed to parse {pdf_path}')


import typing

from elasticsearch import Elasticsearch

from . import settings
from . import parser

client = Elasticsearch(settings.ELASTIC_HOST)

XML_USPTO_POST_2001 = 'xml_uspto_post_2001'


class IndexedPatent(parser.GoogleParsedPatent):
    summary: typing.Optional[str] = None
    short_summary: typing.Optional[str] = None
    topics: typing.Optional[typing.List[str]] = None
    text: typing.Optional[str] = None
    text_format: typing.Optional[str] = None


claim_embedding_scheme = {
    "properties": {
        "patent_doc_id": {
            "type": "keyword"
        },
        "claim_id": {
            "type": "keyword"
        },
        "embedding": {
            "type": "dense_vector",
            "dims": 1024
        },
        'filing_date': {
            'type': 'date'
        },
        'priority_date': {
            'type': 'date'
        }
    }
}


specification_embedding_scheme = {
    "properties": {
        "patent_doc_id": {
            "type": "keyword"
        },
        "paragraph_id": {
            "type": "keyword"
        },
        "embedding": {
            "type": "dense_vector",
            "dims": 1024
        },
        'filing_date': {
            'type': 'date'
        },
        'priority_date': {
            'type': 'date'
        }
    }
}


patent_scheme = {
    "properties": {
        "unique_id": {
            "properties": {
                "patent_number": {
                    "type": "keyword"
                },
                "country_code": {
                    "type": "keyword"
                },
                "kind_code": {
                    "type": "keyword"
                },
            }
        },
        "claims": {
            "type": "text"
        },
        'claims_format': {
            'properties': {
                'data_format': {
                    'type': 'keyword'
                },
                'tree_structure': {
                    'type': 'keyword'
                }
            }
        },
        'specification': {
            "type": "text"
        },
        'specification_format': {
            'properties': {
                'data_format': {
                    'type': 'keyword'
                },
                'tree_structure': {
                    'type': 'keyword'
                }
            }
        },
        'title': {
            'type': 'text'
        },
        'summary': {
            'type': 'text'
        },
        'summary_model_version': {
            'type': 'keyword'
        },
        'short_summary': {
            'type': 'text'
        },
        'short_summary_model_version': {
            'type': 'keyword'
        },
        'topics': {
            'type': 'keyword'
        },
        'topics_model_version': {
            'type': 'keyword'
        },
        'text': {
            'type': 'text'
        },
        'text_format': {
            'type': 'keyword'
        },
        'filing_date': {
            'type': 'date'
        },
        'priority_date': {
            'type': 'date'
        }
    }
}


def create_specification_embedding_index():
    try:
        client.indices.create(index='patents_specification_embedding')
    except Exception as e:
        pass
    client.indices.put_mapping(index='patents_specification_embedding', body=specification_embedding_scheme)


def create_claim_embedding_index():
    try:
        client.indices.create(index='patents_claim_embedding')
    except Exception as e:
        pass
    client.indices.put_mapping(index='patents_claim_embedding', body=claim_embedding_scheme)


def create_patents_index():
    try:
        client.indices.create(index='patents')
    except Exception as e:
        pass
    client.indices.put_mapping(index='patents', body=patent_scheme)


def index_patent(unique_id, text, title, text_format=None, filing_date=None, priority_date=None):
    body = {
        "unique_id": {
            "patent_number": unique_id.patent_number,
            "kind_code": unique_id.kind_code,
            "country_code": unique_id.country_code
        },
        "text": text,
        "title": title
    }

    if text_format:
        body['text_format'] = text_format
    else:
        body['text_format'] = XML_USPTO_POST_2001

    if filing_date:
        body['filing_date'] = filing_date

    if priority_date:
        body['priority_date'] = priority_date
        
    client.index(index='patents', id=str(unique_id), body=body)
    return str(unique_id)


def index_claim_embedding(patent_doc_id, embedding, claim_id=None):
    body = {
        "patent_doc_id": patent_doc_id,
        "embedding": embedding
    }
    if claim_id:
        body['claim_id'] = claim_id

    client.index(index='patents_claim_embedding', body=body)


def index_specification_embedding(patent_doc_id, embedding, paragraph_id=None):
    body = {
        "patent_doc_id": patent_doc_id,
        "embedding": embedding
    }
    if paragraph_id:
        body['paragraph_id'] = paragraph_id
    client.index(index='patents_specification_embedding', body=body)


def get_patent_summary(patent: parser.GoogleParsedPatent) -> str:
    response = client.get(index='patents', id=str(patent.unique_id))
    return response['_source'].get('summary')


def update_patent_by_id(patent: parser.GoogleParsedPatent, **kwargs):
    client.update(index='patents', id=str(patent.unique_id), body={'doc': kwargs})


def patent_exists(patent_unique_id: parser.PatentUniqueID | str) -> bool:
    try:
        client.get(index='patents', id=str(patent_unique_id))
        return True
    except Exception as e:
        return False
    

def get_patent_content_by_id(patent_id: str) -> IndexedPatent | None:
    try:
        response = client.get(index='patents', id=patent_id)
    except Exception as e:
        logger.error('Patent id %s not found in the index', patent_id)
        return None
    return IndexedPatent(**response['_source'])